"""
Standard functions for general data.

Many of these are added to Plate and other classes as methods with
default arguments.
"""

import numpy as np
import pandas as pd
from .parsers import _parse_data_obj
from .util import check_df_col


def normalize(
    obj,
    value=None,
    to=None,
    zero=None,
    groupby=None,
    # devs=False,
    prefix='normalized_',
):
    """Flexible scaling and translation of annotated data to help
    normalize and compare different experiments. Use `value` to choose
    the data to be normalized. Set `zero` to True to use the minimum
    of the data as an explicit 0 value (e.g., no negative values), or
    assign an annotation in a different column to be set to zero. The
    max of the data is set to 1, unless given an annotation to set to a
    value of 1. See kwargs below for more details.

    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame object
        If DataFrame and `value = None`, the final column is used as the
        value.
    value: str or list of str, default None
        Name of column(s) to normalize. If no argument is given, uses
        value_name for Plate objects or the right-most column for 
        DataFrames.
    to: str, default None
        Which group to set as the normal (1) value, following the form
        of the string '{column}={group}'.
    zero: str or bool, default None
        Which group to set as the zero value (if str) or to use the min
        of the data as zero (if False). If str, follows the same form as
        `to`, '{column}={group}'.
    groupby: str, default None
        Name of a column that contains group information, if multiple
        independent normalizations should be performed.
    prefix: str, default 'normalized_'
        Prefix used for the normalized column name(s).

    Examples:
    ---------
    # Normalize a single data column with no arguments
    >>> df = pd.DataFrame({
    ...     'condition': [1, 1, 1, 1, 2, 2, 2, 2],
    ...     'value': [0.98, 1.02, 1.07, 0.95, 0.33, 0.20, 0.25, 0.27]
    ... })
    >>> ns.normalize(df)
        condition    value    normalized_value
    0   1            0.98     0.915888
    1   1            1.02     0.953271 
    2   1            1.07     1.000000
    3   1            0.95     0.887850
    4   2            0.33     0.308411
    5   2            0.20     0.18916
    6   2            0.25     0.233645
    7   2            0.27     0.252336

    # Condition 1 is a positive control, condition 2 is a negative.
    >>> ns.normalize(
    ...     df,
    ...     to='condition=1',
    ...     zero='condition=2',
    ... )
        condition    value    normalized_value
    0   1            0.98     0.966330
    1   1            1.02     1.020202 
    2   1            1.07     1.087542
    3   1            0.95     0.925926
    4   2            0.33     0.090909
    5   2            0.20    -0.084175
    6   2            0.25    -0.016835
    7   2            0.27     0.010101
    """
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)

    # Get value list ready
    if value is None:
        value = auto_value
    if not isinstance(value, list):
        values = [value]
    else:
        values = value

    # Set up groups
    if not isinstance(groupby, (tuple, list)):
        groupby = [groupby]

    for group in groupby:
        check_df_col(df, group, name='groupby')

    # Group or create fake groupby object
    unique_dfs = (df.groupby(groupby) if groupby != [None]
                    else [(None, df.copy())])

    df_list = []
    # Iterate through each dataframe group
    for name, sub_df in unique_dfs:

        sub_df = sub_df.copy()

        # Iterate through each value
        for value in values:
            check_df_col(sub_df, value, name='value')

            norm_string = f'{prefix}{value}'

            # Set the zero val
            if zero == True:
                zero_val = sub_df[value].min()
            elif isinstance(zero, str):
                split = zero.split('=')
                if len(split) != 2:
                    raise ValueError(
                        f"'zero' value specified incorrectly, must be of the "\
                        "form '{column}={group}'."
                    )

                col, val = split

                # Check that col in columns
                check_df_col(sub_df, col, name='column')

                # Check that the given value is found in the column
                if val not in [str(i) for i in sub_df[col]]:
                    raise ValueError(
                        f"The value '{val}' is not a value in the column '{col}'."
                    )
                subset = [val == str(i) for i in sub_df[col]]
                zero_val = sub_df[subset][value].mean()
            elif zero is None or zero == False:
                zero_val = 0
            else:
                raise TypeError(
                    f"Type of 'zero' argument is incorrectly specified. Must be bool or string.")

            sub_df[norm_string] = sub_df[value] - zero_val

            # Set the one val
            if to is None:
                one_val = sub_df[norm_string].max()
            elif isinstance(to, str):
                split = to.split('=')
                if len(split) != 2:
                    raise ValueError(
                        f"'to' value specified incorrectly, must be of the "\
                        "form '{column}={value}'."
                    )

                col, val = split

                # Check that col in columns
                check_df_col(sub_df, col, name='column')

                # Check that the given value is found in the column
                if val not in [str(i) for i in sub_df[col]]:
                    raise ValueError(
                        f"The value '{val}' is not a value in the column '{col}'."
                    )
                subset = [val == str(i) for i in sub_df[col]]
                one_val = sub_df[subset][norm_string].mean()
            else:
                raise TypeError(
                    f"Type of 'to' argument is incorrectly specified. Must be bool or string.")

            sub_df[norm_string] = sub_df[norm_string] / one_val

        # After adding normalized values, store in df_list
        df_list.append(sub_df)

    # Once finished, concat together
    # Note: the index is preserved, and returns the data in the same
    # order as the input, even with groupby.
    return_df = pd.concat(df_list)

    return return_df


def aggregate_replicates(
    obj,
    variable=None,
    value=None,
    groupby=None,
    prefix='mean_',
):
    """Checks for the presence of replicates in the values of a dataset,
    given some experimental conditions. Returns True if the number of
    rows assigned to the given aggregation columns (variable+groupby)
    is greater than 1, indicating that replicates were performed.
    
    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame object
        If DataFrame and `value = None`, the final column is used as the
        value.
    variable: str, default None
        Name of column of data frame for the independent variable,
        indicating a specific experimental condition.
    value: str or list of str, default None
        Name of column(s) to normalize. 
    value: str, default None
        Name of column of data frame for the dependent variable,
        indicating an experimental observation. If no argument is given,
        uses value_name for Plate objects or the right-most column for 
        DataFrames.
    groupby: str or list of str, default None
        Column name or list of column names that indicates how the
        data set should be split.
    prefix: str, default 'mean_'
        The prefix for the new column of aggreagated values, if
        replicates is True.
    
    Returns:
    --------
    replicates: bool
        True if replicates are present.
    df: pd.DataFrame
        The DataFrame containing averaged 'variable'+'groupby' values,
        if replicates is True. Otherwise returns the original DataFrame.
        The aggregated data is in a new column named '{prefix}{value}'.

    Examples:
    ---------
    # Try to aggregate with 8 different variables (just ID)
    # returns False, input df
    >>> df = pd.DataFrame({
    ...     'id': [1, 2, 3, 4, 5, 6, 7, 8],
    ...     'condition': [1, 1, 1, 1, 2, 2, 2, 2],
    ...     'value': [0.98, 1.02, 1.07, 0.95, 0.33, 0.20, 0.25, 0.27]
    ... })
    >>> ns.aggregate_replicates(df, variable='id')
    (False,
        id      condition    value
    0   1       1            0.98
    1   2       1            1.02
    2   3       1            1.07
    3   4       1            0.95
    4   5       2            0.33
    5   6       2            0.20
    6   7       2            0.25
    7   8       2            0.27)

    # Aggreagte on 'condition', which contains two groups
    # returns True, df iwth aggregated column
    >>> ns.aggregate_replicates(df, variable='condition')
    (True,
        id      condition    value    mean_value
    0   1       1            0.98     1.0050
    1   2       1            1.02     1.0050
    2   3       1            1.07     1.0050
    3   4       1            0.95     1.0050
    4   5       2            0.33     0.2625
    5   6       2            0.20     0.2625
    6   7       2            0.25     0.2625
    7   8       2            0.27     0.2625

    # Aggregate on two groups, just passing groupby
    # returns True, aggregated df
    >>> df = pd.DataFrame({
    ...     'condition_1': [1, 1, 1, 1, 2, 2, 2, 2],
    ...     'condition_2': [True, True, False, False]*2,
    ...     'value': [0.98, 1.02, 1.07, 0.95, 0.33, 0.20, 0.25, 0.27]
    ... })
    >>> ns.aggregate_replicates(
    ...     df,
    ...     groupby=['condition_1', 'condition_2]
    ... )
    (True,
        condition_1 condition_2    value    mean_value
    1   1           True           1.02     1.000
    0   1           True           0.98     1.000
    2   1           False          1.07     1.010
    3   1           False          0.95     1.0010
    4   2           True           0.33     0.265
    5   2           True           0.20     0.265
    6   2           False          0.25     0.260
    7   2           False          0.27     0.260
    """
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)

    # Get value list ready
    if value is None:
        value = auto_value
    
    # Unpack the experimental conditions into a single list of arguments
    if not isinstance(groupby, (list, tuple)):
        groupby = [groupby]
    args = [elem for elem in (variable, *groupby) if elem is not None]

    # Get row counts to determine if there are replicates
    grouped = df.groupby(args)[value]
    counts = grouped.count()
    replicates = bool([True for count in counts if count > 1])

    # Average the values and return
    if replicates:
        df_mean = grouped.mean().reset_index()
        df_mean.columns = list(df_mean.columns[:-1]) + [f'{prefix}{value}']
        df = df.merge(df_mean)

    return replicates, df



"""
Tools for parsing and regularizing data found in Plate objects.
"""

import pandas as pd
from itertools import product


def _infer_casing(string):
    """Guesses the case of a string."""
    # Acceptable cases
    cases = ['lower', 'title', 'capitalize', 'upper']

    found_cases = [case for case in cases
                   if getattr(str, case)(string) == string]

    return found_cases

def _infer_padding(well):
    """Guesses if a well is padded (A01) or not (A1). Returns False
    if it cannot be guessed (on double-digit column).
    """
    # Assume False
    padded = False

    row = well[0]
    str_col = well[1:]
    int_col = str(int(str_col))

    # Return True is str form != int form
    if len(str_col) != len(int_col):
        padded = True

    return padded


def _get_ordered_locs(df):
    """Returns location information in 'well', 'row', 'column' order."""
    well = [val for val in df.columns if val.lower() == 'well']
    row = [val for val in df.columns if val.lower() == 'row']
    column = [val for val in df.columns if val.lower() == 'column']

    for loc in (well, row, column):
        if not loc:
            raise ValueError(
                f'Could not find column for "{loc}"'
            )
        if len(loc) > 1:
            raise ValueError(
                f'Too many columns found: "{loc}"'
            )

    return well[0], row[0], column[0]


def _parse_data_obj(obj):
    """Gets necessary information for 96-well-based plots"""
    if isinstance(obj, pd.DataFrame):
        df = obj.copy()
        value_name = df.columns[-1]

        # Determine case
        lowers = [True for val in df.columns if val.lower() == val]
        caps = [True for val in df.columns if val.capitalize() == val]
        case = str.lower if lowers > caps else str.capitalize

    # Assume Plate
    else:
        try:
            df = obj.df.copy()
            value_name = obj.value_name
            case = obj._set_case
        except AttributeError:
            raise ValueError(
                'Data format not recognized, only pandas.DataFrame and '
                'ninetysix.Plate objects accepted.'
            )

    # Get well, row, and column values
    try:
        well, row, col = _get_ordered_locs(df)
        locs = (well, row, col)
    except ValueError:
        locs = None

    return df, locs, value_name, case

def pad(well, padded=True):
    """Converts to or from zero-padded column names
    (e.g., A1 <-> A01)
    """
    # Get row and column
    row = well[0]
    col = well[1:]

    # Convert to int then to str (gives non-padded)
    col = str(int(col))

    # Pad
    if padded:
        if len(col) == 1:
            col = '0'+col

    well = row+col
    
    return well

def well_regex(input_dict, padded=None):
    """Parses simple regex-like syntax for well keys in a
    dictionary and expands it. Accepts up to 396-well plate
    nomenclature (A–P, 1–24). If no regex-like structure is
    detected, will return the same dictionary.

    Parameters
    ----------
    input_dict : dictionary
        A dictionary potentially containing keys of the form
        '[A-E][1-10]', which would expand to 50 keys of the form
        'A1', 'A2', ... , 'E9', 'E10', each with the same value as
        the original key.
    padded : bool, default None
        Whether or not to zero pad the wells, i.e., return keys of
        the form 'A01', 'A02', ... , 'E09', 'E10'. Will occur
        regardless of the regex key form (both '[A-E][1-10]' and
        '[A-E][01-10]' will return A1 if padded=False, else A01.
        If None, infers padding from inputs if possible. 
    """
    # Set up the new dict
    parsed_dict = input_dict.copy()

    # Acceptable values to search
    rows = list('ABCDEFGHJLMNOP')
    cols = [str(i) for i in range(1, 25)]

    # Set up a dictionary that can change during the loop
    working_dict = parsed_dict.copy()
    for assignment in working_dict.keys():

        # Pop the assignment and store
        value = parsed_dict.pop(assignment)
        
        # If regex-ed
        if '[' in assignment:

            # Determine if rows, columns, or both are regex-ed
            row = True if assignment[0] == '[' else False
            col = True if assignment[-1] == ']' else False
            
            if row:
                matching_rows = []
                
                # Get the row regex contents
                row_vals = assignment[1:assignment.find(']')]
                
                # Iterate through each group, if given
                comma_split = row_vals.split(',')
                for row_set in comma_split:
                    
                    # Get the range
                    hyphen_split = row_set.split('-')
                    
                    # Append range to list
                    if len(hyphen_split) > 1:
                        row_range = rows[rows.index(hyphen_split[0]):rows.index(hyphen_split[1])+1]
                        matching_rows += row_range
                        
                    # Else append single value to list
                    else:
                        matching_rows += hyphen_split

            else:
                matching_rows = assignment[0]
                
            if col:
                matching_cols = []
                
                # Get the column regex contents
                col_vals = assignment[assignment.rfind('[')+1:-1]
                
                # Iterate through each group, if given
                comma_split = col_vals.split(',')
                for col_set in comma_split:
                    
                    # Get the range
                    hyphen_split = [val for val in col_set.split('-')]
                    
                    # Append range to list
                    if len(hyphen_split) > 1:
                        padding = set([_infer_padding(('_'+val))
                                    for val in hyphen_split
                                    if int(val) < 10])

                        if len(padding) > 1:
                            raise ValueError(
                            f'Inconsistent zero padding in "{col_vals}"; '\
                            'use the same form for best results.'
                        )

                        if all(padding):
                            cols = [pad('_'+col)[1:] for col in cols]

                        start = cols.index(hyphen_split[0])
                        end = cols.index(hyphen_split[1])+1

                        col_range = cols[start:end]
                        matching_cols += col_range
                    
                    # Else append single value to list
                    else:
                        matching_cols += hyphen_split

            else:
                if row:
                    col_start = assignment.find(']')
                else:
                    col_start = 0
                
                matching_cols = [assignment[col_start+1:]]

            wells = [''.join(well)
                for well in product(matching_rows, matching_cols)]

        # Non-regex
        else:
            wells = [assignment]
        
        for well in wells:
            parsed_dict[well] = value

    # Infer padding
    if padded is None:
        padded = [True for well in parsed_dict
                   if _infer_padding(well)]
    # Pad
    if padded:
        padded_dict = parsed_dict.copy()
        for well in parsed_dict:
            value = padded_dict.pop(well)
            try:
                padded_dict[pad(well)] = value
            except ValueError:
                padded_dict[well] = value
    
        parsed_dict = padded_dict

    return parsed_dict



"""
Plates
"""

import numpy as np
import pandas as pd

from .plate import Plate

class Plates(Plate):
    """Class for handling multiple Plate-like objects"""
    
    def __init__(
        self,
        plate_list,
        plate_name=None,
        name_list=None,
        # Plate args if non-plate obj is passed
        locations=None,
        values=None,
        case=None,
        zero_padding=None,
    ):
        # Assign attributes
        self.plate_list = plate_list
        self.plate_name = plate_name
        self.name_list = name_list
        self._locations = locations
        self._values = values
        self._case = case
        self._zero_padding = zero_padding

        self = self._generate_plates()

    def _generate_plates(self):
        # Get type of plate_list passed
        ziplike = (True if isinstance(self.plate_list[0], (list, tuple))
                        else False)

        # If list of plate-like objs: [obj, obj, ...]
        if not ziplike:
            # Assign column name
            if self.plate_name is None:
                self.plate_name = 'plate'

                # Assign the plate name values for each plate
                if self.name_list is None:
                    self.name_list = range(1, len(self.plate_list)+1)
            # else:
                # for plate in self.plate_name:
                #     try:
                #         plate[plate_name]
                #     except KeyError:
                #         raise ValueError(
                #             f'{plate_name} not found in your columns, options'\
                #             f' are {plate.columns}.'
                #         )
        
        # If list of plate-name pairs: [(obj, 1), (obj, 2), ...]
        elif ziplike:
            # Unzip
            self.plate_list, self.name_list = (
                [plate for (plate, name) in self.plate_list],
                [name for (plate, name) in self.plate_list]
            )

            # Assign column name
            if self.plate_name is None:
                self.plate_name = 'plate'

        # Determine if Plate or DataFrame
        df_bool = [True for plate in self.plate_list
                      if isinstance(plate, pd.DataFrame)]
        if sum(df_bool) == len(self.plate_list):
            df_bool = True
        elif not df_bool:
            df_bool = False
        else:
            raise ValueError(
                'Mixed object types are not allowed in Plates'
            )
        
        # Make plates
        if df_bool:
            plates = []
            for df in self.plate_list:
                plate = Plate(
                    df,
                    value_name=self._value_name,
                    zero_padding=self._zero_padding,
                    case=self._case
                )
                plate.locations = self._locations
                plate.values = self._values
                plates.append(plates)
            
            self.plate_list = plates

        # Check that all objects are the same
        locations = [plate.locations for plate in self.plate_list]
        annotations = [plate.annotations for plate in self.plate_list]
        values = [plate.values for plate in self.plate_list]
        for plate_attr in (locations, annotations, values):
            differences = set(plate_attr[0]).difference(*plate_attr[1:])
            if differences:
                raise ValueError(
                    'Plate objects contain unmatched data columns'
                )

        

        # Add plate name data
        # if self.name_list is None:
        #     for plate in self.plate_list: